/*
 * (c) 1996-2000 by Polar Pyramid.
 */

/*
 * log.c
 * logging with variable nr of arguments.
 * all log entries go to screen and file.
 * automatically identifies the line of sourcecode for each log entry.
 *
 * 960709 BSt
 *   Added logging to game console.
 * 970315 BSt
 *   Adjusted log margins, so it will handle 8.2 source filenames.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/stat.h>	// for umask()


#include "auxil.h"

const char * const EOLN = "\n";

static logfunc_t extra_logger = 0;
static FILE *auxil_log_file = 0;
static std::string auxil_log_filename("nb.log");
static bool auxil_stdout_write = 0;
static bool log_enabled = false;
static int indent_level = 0;

/*---------------------------------------------------------------------------*/
/*
  auxil_log messes up profile runs, because it's slow. We need to
  disable it during profiling. 
  */
 
void auxil_log_enable(bool b)
{
  // logging messes up profiling so ignore turn on

  if (b && getenv("NBPROFILE") && (!strcmp(getenv("NBPROFILE"),"Y")))
    return;

  log_enabled = b;
}

/*---------------------------------------------------------------------------*/

void auxil_log_stdout_write(bool enable)
{
  auxil_stdout_write = enable;
}

/*---------------------------------------------------------------------------*/

void auxil_log_extrafunc(logfunc_t f) 
{
  extra_logger = f;
}

/*---------------------------------------------------------------------------*/

void auxil_log_set_filename(const std::string &filename)
{
  if (auxil_log_file)
  {
    fclose(auxil_log_file);
    auxil_log_file = 0;
  }
  std::string lp = GetLogPath();
  auxil_log_filename = lp + filename;
}

// helper func. replaces GNU basename()
const char *auxil_strip_dir(const char *filename)
{
  int len = strlen(filename);
  if (!len) return filename;
  const char *p = filename + len - 1;
  while (*p && p>filename && *p != '/')
    p--;
  if (*p=='/') p++;
  return p;
}

/*---------------------------------------------------------------------------*/
/*
 * auxil_log
 *
 * This is a var-args function.
 * logs of several types can be made: INF, WRN, ERR, FAT.
 */

void auxil_log(
  int   logtype,
  char* filename,
  int   linenr,
  const char* format,
  ...
) 
{
  if (!log_enabled)
    return;

  if (!auxil_log_file) 
  {
    umask(0);
    auxil_log_file = fopen(auxil_log_filename.c_str(), "wb");
    if (!auxil_log_file)
    {
      fprintf(stderr,"Cannot open log file '%s'!\n", auxil_log_filename.c_str());
      exit(1);
    }
  }

  char          where[128];  /* place in source code */
  sprintf(where,"%-12s(%4d) ", auxil_strip_dir(filename), linenr);

  char          *what;       /* warning/error/fatal/info */
  what="WRONG LOGTYPE:";
  switch(logtype) 
  {
  case LOGID_INF: what=""; break;
  case LOGID_WRN: what="WARNING:"; break;
  case LOGID_ERR: what="ERROR:"; break;
  case LOGID_FAT: what="FATAL ERROR:"; break;
  case LOGID_TRM: what="NORMAL PROGRAM TERMINATION:"; break;
  }

  int           i;
  char          indent[80];   /* our indentation */
  strcpy(indent,"");
  for (i=0; i<indent_level; i++)
    strcat(indent,"  ");

  char          msg[512];    /* the mesg we are gonna print */
  strcpy(msg,"");

  va_list       ptr;         /* our argument list of variable lenght */
  va_start(ptr,format);
  vsprintf(msg, format, ptr);
  va_end(ptr);

  char          line[512];   /* the line we are gonna print */
  strcpy(line, where);
  strcat(line, indent);
  strcat(line, what);
  strcat(line, msg);

  if (auxil_stdout_write)
  {
    fprintf(stdout, "%s%s", line, EOLN); 
    fflush(stdout);
  }
  if (auxil_log_file)
  { 
    fprintf(auxil_log_file, "%s%s", line, EOLN); 
    fflush(auxil_log_file); 
  }

  if (extra_logger && ((*line)!='-'))
    extra_logger(line);

#ifdef OUTPUT_DBG_STR_NB
  strcat(line, EOLN); 
  OutputDebugString(line);  
#endif  

  // I want to see errors, warnings and such always
  // the log file is not always checked - Johan

  if (logtype != LOGID_INF)
    puts(line);

  if (logtype == LOGID_FAT)
    exit(0); 
}

void auxil_log_indent_up(void) 
{ 
  auxil_log(LOG_INF,"{");
  indent_level++; 
}

void auxil_log_indent_down(void) 
{ 
  if (indent_level) 
    indent_level--;
  auxil_log(LOG_INF,"}");
}

