/*
 * Tux vs Clippy
 * (c)2002-2003 by Bram Stolk
 * b.stolk at chello.nl
 * Licensed under GPL
 */

#include <math.h>
#include <errno.h>
#include <fcntl.h>
#include <linux/joystick.h>
#include <sys/ioctl.h>	// for ioctl()
#include <unistd.h>	// for close()
#include <stdio.h>	// for perror()

#include <algorithm>


#include "auxil.h"
#include "inppad.h"


InputGamePad::InputGamePad(const std::string &devfile) :
fd(0),
DeadMargin(0.12)
{
  fd = open(devfile.c_str(), O_RDONLY | O_NONBLOCK);
  if (fd==-1)
  {
    perror("open() on /dev/joystick failed");
    auxil_log(LOG_ERR,"No joystick");
    return;
  }
  opened = true;

  int retval;

  int version;
  retval = ioctl(fd, JSIOCGVERSION, &version);
  if (retval == -1)
    perror("ioctl JSIOCGVERSION failed");

  unsigned char lo = version;
  unsigned char md = (version>>8);
  unsigned char hi = (version>>16);

  retval = ioctl(fd, JSIOCGAXES, &axiscount);
  if (retval == -1)
    perror("ioctl JSIOCGAXES failed");

  retval = ioctl(fd, JSIOCGBUTTONS, &buttoncount);
  if (retval == -1)
    perror("ioctl JSIOCGBUTTONS failed");

  char name[128];
  retval = ioctl(fd, JSIOCGNAME(128), name);
  if (retval == -1)
    perror("ioctl JSIOCGNAME failed");

  auxil_log(LOG_INF,"%s", name);
  auxil_log(LOG_INF,"driver version: %d.%d.%d", hi,md,lo);
  auxil_log(LOG_INF,"axiscount=%d, buttoncount=%d", axiscount, buttoncount);

  int i;
  for (i=0; i<axiscount; i++)
    AxisValues.push_back(32768);
  for (i=0; i<buttoncount; i++)
    ButtonValues.push_back(false);
}


InputGamePad::~InputGamePad()
{
  if (opened)
  {
    auxil_log(LOG_INF,"Closing down joystick device");
    int retval = close(fd);
    if (retval==-1)
      perror("close() failed");
  }
}


void InputGamePad::Update(float dt)
{
  struct js_event ev;
  int retval;
  do
  {
    retval = read(fd, &ev, sizeof(ev));
    if (retval == -1)
    {
      if (errno != EAGAIN)
      {
        perror("read() on joystick failed");
        auxil_log(LOG_FAT,"joystick read failure");
      }
    }
#if 0
    int type = ev.type & (~JS_EVENT_INIT);
#else
    int type = ev.type;
#endif
    switch(type)
    {
      case JS_EVENT_BUTTON:
        if (ev.number < ButtonValues.size())
          ButtonValues[ev.number] = ev.value;
        else
          auxil_log(LOG_ERR,"Illegal buttonnr %d", ev.number);
        break;
      case JS_EVENT_AXIS:
        if (ev.number < AxisValues.size())
          AxisValues[ev.number] = ev.value;
        else
          auxil_log(LOG_ERR,"Illegal axisnr %d", ev.number);
        break;
      default:
//        puts("????");
        break;
    }
  } while (retval>0);
}


float InputGamePad::GetAxisValue(int nr) const
{
  assert(nr < AxisValues.size());

  float x=AxisValues[nr]/32768.0;
  if (fabsf(x)<DeadMargin) 
    x=0;
  else
    x -= (x>0)?DeadMargin:-DeadMargin;

  return x / (1.0 - DeadMargin);
}


bool InputGamePad::GetButtonValue(int nr) const
{
  assert(nr < ButtonValues.size());
  return ButtonValues[nr];
}

