/*
 * Tux vs Clippy
 * (c)2002-2003 by Bram Stolk
 * b.stolk at chello.nl
 * Licensed under GPL
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>

// include autoconf/automake's configuration
#include "config.h"

// auxil
#include "timer.h"

// rot
#include "setup.h"
#include "rot.h"
#include "world.h"

#include "im16.h"
#include "im32.h"
#include "framebuffer32.h"
#include "framebufferdga.h"
#include "framebuffersdl.h"
#include "sprite.h"
#include "gfxscreen.h"

// pres
#include "presdb.h"
#include "bullet.h"
#include "player.h"
#include "enemy.h"
#include "generator.h"
#include "shield.h"
#include "scorecounter.h"

// inp
#include "inppad.h"
#include "inpkey.h"
#include "inpsdl.h"

// snd
#include "soundengine.h"

// main
#include "numbercloud.h"



// The command line options

static int         opt_accellaxis=1;
static int         opt_cancelbutton=2;
static std::string opt_vo="sdl";
static std::string opt_ao="sdl";
static bool        opt_reverseaccel=false;


static bool ParseOption(char *o)
{
  if (!strncmp(o,"vo=",3))
  {
    opt_vo = o+3;
    return true;
  }
  if (!strncmp(o,"ao=",3))
  {
    opt_ao = o+3;
    return true;
  }
  if (!strncmp(o,"accellaxis=",11))
  {
    opt_accellaxis = atoi(o+11);
    return true;
  }
  if (!strncmp(o,"cancelbutton=",13))
  {
    opt_cancelbutton = atoi(o+13);
    return true;
  }
  if (!strcmp(o,"reverseaccel"))
  {
    opt_reverseaccel=true;
    return true;
  }
  return false;
}



static bool PlayLevel
(
  int lvlnr, 
  FrameBuffer *fbuf,
  InputDevice *pad,
  int alliance
)
{
  Image32 im(GetImagePath()+"console.tif");
  if (fbuf->bpp==32 || fbuf->bpp==24)
    im.SwapRedBlue();
  fbuf->Blit(im,0,0);

  GfxScreen &scn = GfxScreen::TheOne();
  PresenceDB &pdb = PresenceDB::TheOne();
  ScoreCounter &score = ScoreCounter::TheOne();

  Timer timer;

  World wld((lvlnr & 1)?"void":"street");
  char mapname[80];
  sprintf(mapname,"map%d",lvlnr);
  wld.CreateWorld(mapname);
  const Image16 &rot_im = wld.GetRotationBuffer();
  Image32 ply_im32(PSCREENW, PSCREENH);
  Image16 ply_im16(PSCREENW, PSCREENH);
  short *rotbuf = rot_im.data;

  // Add a player
  auxil_log(LOG_INF,"Creating player");
  Player player("player", (alliance==0)?"penguinwalk":"paperclipwalk", wld);
  pdb.Clear();
  pdb.Add(&player);

  auxil_log(LOG_INF,"Populating world");
  // Add Generators
  if (alliance==0)
  {
    Generator *generator0 = new Generator("generator","icongates","paperclip",wld);
    generator0->pos = wld.GetRandomWorldPosition();
    generator0->SetBalloonName("fud");
    pdb.Add(generator0);
    Generator *generator1 = new Generator("generator","iconsteve","paperclip",wld);
    generator1->pos = wld.GetRandomWorldPosition();
    generator1->SetBalloonName("developers");
    pdb.Add(generator1);
  }
  else
  {
    Generator *generator0 = new Generator("generator","iconrms","penguin",wld);
    generator0->pos = wld.GetRandomWorldPosition();
    generator0->SetBalloonName("gnu");
    pdb.Add(generator0);
    Generator *generator1 = new Generator("generator","iconlinus","penguin",wld);
    generator1->pos = wld.GetRandomWorldPosition();
    generator1->SetBalloonName("domination");
    pdb.Add(generator1);
  }

  // Add Shields
  for (int shieldnr=0; shieldnr<2; shieldnr++)
  {
    Shield *shield = new Shield("shield", "swirl", wld);
    shield->pos = wld.GetRandomWorldPosition();
    pdb.Add(shield);
  }

  // Add Enemies
  for (int enr=0; enr<30+lvlnr*5; enr++)
  {
    Vector3 epos = wld.GetRandomWorldPosition();
    Enemy *enemy = new Enemy("enemy", (alliance==0)?"paperclip":"penguin", wld);
    enemy->pos = epos;
    pdb.Add(enemy);
  }

  // add lvl indicator
  NumberCloud::TheOne().Fire(wld,lvlnr+1);

  int framecount=0;
  bool weapon_loaded = false;

  auxil_log(LOG_INF,"Entering mainloop of PlayLevel");
  while(!player.dead && !pad->GetButtonValue(opt_cancelbutton))
  {
    float dt = timer.SecondsElapsed();
    timer.Start();
    //printf("dt=%f\n", dt);
    if (dt < 0.001)
      dt = 0.001;
    if (dt > 0.5)
      dt = 0.5;

    wld.RenderRotationBuffer(player.pos);
    float playerangle = get_angle_from_2d_heading(player.dir);
    scn.SetPivot(player.pos, playerangle - 0.5*M_PI);
    scn.RotateBuffer(rotbuf);

    pdb.Render(scn);
    score.Update(dt);
    score.Render(scn);

    if (fbuf->bpp==32 || fbuf->bpp==24)
    {
      scn.GetImage().Crop(ply_im32, MAXSPRITEW, MAXSPRITEH);
      ply_im32.SwapRedBlue();
      fbuf->Blit(ply_im32, 80, 48, false);
    }
    else
    {
      scn.GetImage().Crop(ply_im16, MAXSPRITEW, MAXSPRITEH);
      fbuf->Blit(ply_im16, 80, 48, false);
    }

    pad->Update(dt);

    // Update player orientation
    float pan = pad->GetAxisValue(0);
    Vector2 newdir;
    RotateVector2WithRealAngle(0.09*pan, player.dir, newdir);
    newdir.normalize();
    player.dir=newdir;
    player.ori=newdir;

    float accel = pad->GetAxisValue(opt_accellaxis);
    if (opt_reverseaccel)
      accel = -accel;
    if (accel < 0)
      player.vel += 30*accel;
    else
      player.vel += 10*accel;

    bool buta=pad->GetButtonValue(0);
    bool butx=pad->GetButtonValue(1);
    if (!buta && !butx)
    {
      weapon_loaded=true;
    }
    if ((weapon_loaded && buta) || (!(framecount&1) && butx))
    {
      static int channr=0;
      Bullet *b = new Bullet("iconx", wld);
      b->dir = player.dir;
      b->pos = player.pos + (player.radius+b->radius) * player.dir;
      b->vel = player.vel + 300.0;
      b->angularvel = (rand()&1)?0.5:-0.5;
      pdb.Add(b);
      SoundEngine::TheOne().Play("fire",1+channr,0);
      channr = (channr)?0:1;
      weapon_loaded = false;
    }
    pdb.Update(dt);

    if (pdb.Count(typeid(Enemy))==0)
      return true;

    framecount++;
  }
  return false;
}


static int SelectAlliance(FrameBuffer *fbuf, InputDevice *pad)
{
  Image32 im0(GetImagePath()+"bout0000.tif");
  Image32 im1(GetImagePath()+"bout0001.tif");
  if (fbuf->bpp==32 || fbuf->bpp==24)
  {
    im0.SwapRedBlue();
    im1.SwapRedBlue();
  }

  fbuf->Blit(im0,0,0);

  int alliance = 0;
  while (! pad->GetButtonValue(0))
  {
    if (pad->GetAxisValue(0) < -0.75 && alliance==1)
    {
      alliance=0;
      fbuf->Blit(im0,0,0);
    }
    if (pad->GetAxisValue(0) >  0.75 && alliance==0)
    {
      alliance=1;
      fbuf->Blit(im1,0,0);
    }
    pad->Update(0.05);
    if (pad->GetButtonValue(opt_cancelbutton))
      return -1;
  }
  fbuf->Clear();
  return alliance;
}


int main(int argc, char *argv[])
{
  for (int i=1; i<argc; i++)
  {
    bool ok = ParseOption(argv[i]);
    if (!ok)
    {
      fprintf(stderr,"Usage: %s vo=[dga|fb|sdl] ao=[sdl|oss] accellaxis=n cancelbutton=n reverseaccel\n", argv[0]);
      exit(1);
    }
  }

  auxil_log_set_filename("tuxvsclippy.log");
  auxil_log_enable(true);

  auxil_log(LOG_INF,"Tux vs Clippy version %s", VERSION);
  auxil_log(LOG_INF,"(c)2002-2003 by Bram Stolk. b.stolk at chello.nl");

  auxil_log(LOG_INF,"sprbuf: %dx%d", SPRW, SPRH);
  auxil_log(LOG_INF,"rotbuf: %dx%d", ROTW, ROTH);

  FrameBuffer *fbuf = 0;

  assert(opt_vo == "dga" || opt_vo == "fb" || opt_vo == "sdl");

  if (opt_vo=="dga")
  {
    FrameBufferDGA *fbufdga = new FrameBufferDGA;
    fbufdga->Enable(true);
    fbufdga->Clear();
    fbuf = fbufdga;
  }
  if (opt_vo=="sdl")
  {
    fbuf = new FrameBufferSDL;
  }
  if (opt_vo=="fb")
  {
    fbuf = new FrameBuffer32;
  }

  InputDevice *inpdev=0;

  InputGamePad pad("/dev/js0");
//  InputKeyBoard kbd;
  InputSDL kbd;

  if (pad.opened)
    inpdev = &pad;
  else
    inpdev = &kbd;

  ScoreCounter &score = ScoreCounter::TheOne();

  SoundEngine::SetType(opt_ao);

  bool quit=false;
  while(!quit)
  {
    score.Reset();
    int alliance = SelectAlliance(fbuf, inpdev);
    if (alliance == -1)
    {
      quit=1;
      break;
    }
    int level=0;
    if (getenv("NBLVL"))
      level=atoi(getenv("NBLVL"));

    bool succeeded;
    do
    {
      succeeded = PlayLevel(level, fbuf, inpdev, alliance);
      level = (level+1) % 5;
      std::string plate;
      if (alliance==0)
        plate = (succeeded)?"tuxvictor":"tuxdies";
      else
        plate = (succeeded)?"clippyvictor":"clippydies";
      Image32 im(GetImagePath()+plate+".tif");
      if (fbuf->bpp==32 || fbuf->bpp==24)
        im.SwapRedBlue();
      fbuf->Blit(im,0,0);
      sleep(2);
    } while(succeeded);
  }

  auxil_log(LOG_INF,"Deleting framebuffer");
  delete fbuf;
}

