/*
 * (c) 1996, 1997, 1998 by Polar Pyramid.
 */

// Description:
//	General utility functions
//
// History:
//	Jun 17, 1996:	creation by J. Stolk
//                      parts taken from Graphics Gems gm lib
//                      and some of my older stuff (the templates)
//	Sept 26, 1996:	renamed to MSDEV extensions JSt
//

#ifndef UTILS_H
#define UTILS_H

#include <stdlib.h>

#include <math.h>
#include "const.h"

/////////////////////////////////////////////////////////////////////////////
//
// Random() : return a random double between 0.0 and 1.0
//

inline double Random(void)
{
  return (((double) rand())/RAND_MAX);
}

// Frand(a, b) : geeft random double tussen a en b 

inline double Frand(double a, double b)
{
  return (a + (b - a)*Random());
}

// Irand(a, b) : geeft random int tussen a en b 

inline int Irand(int a, int b)
{
  return a + (int) ((b - a)*Random());
}

//
// SnapValue
//

inline float SnapValue(float    value,
					   float    grid_spacing)
{
  if (value > 0)
  {
	float div = value/grid_spacing;
	return value - grid_spacing*(div - (static_cast<int>(div + 0.5f)));
  }
  else
  {
	float div = -value/grid_spacing;
	return -(-value - grid_spacing*(div - (static_cast<int>(div + 0.5f))));
  }
}

//
// SnapValue
//

inline float SnapValue(float    value,
					   int      grid_spacing)
{
  return SnapValue(value, static_cast<float>(grid_spacing));
}

/////////////////////////////////////////////////////////////////////////////
//
// Safety functions 
// 
// to avoid illegal values for certain math functions which would cause SIGFPE
//

inline double Sqrt(double x)
{
  return (x <= 0.0) ? 0.0 : sqrt(x);
}

// acos : clip illegal values to prevent floatingpoint exceptions

inline double Acos(double x)
{
  return (x <= -1.0) ? acos(-1.0) :(x >= 1.0) ? acos(1.0) : acos(x);
}

inline double Asin(double x)
{
  return (x <= -1.0) ? asin(-1.0) :(x >= 1.0) ? asin(1.0) : asin(x);
}

///////////////////////////////////////////////////////////////////////////////
//
// Template functions 
//
//  useful for float, double, short, int, etc.
//

/////////////////////////////////////////////////////////////////////////////
//
// Sqr(a) : return the square of a
//

template <class TYPE>
inline TYPE Sqr(TYPE a)
{
  return a*a;
}
/////////////////////////////////////////////////////////////////////////////
//
// Abs(a) : return the absolute value ( > 0) of a
//

template <class TYPE>
inline TYPE Abs(TYPE a)
{
  if (a >= 0.0)
    return a;
  return -a;
}
/////////////////////////////////////////////////////////////////////////////
// 
// Swap(a, b) : swap a and b
//

template <class TYPE>
inline void Swap(TYPE &a, TYPE &b)
{
  TYPE tmp = a;
  a = b;
  b = tmp;
}

/////////////////////////////////////////////////////////////////////////////
//
// Max(a, b) : return the maximum of a and b
//

template <class TYPE>
inline TYPE Max(TYPE a, TYPE b)
{
  if (a > b)
    return a;
  return b; 
}

/////////////////////////////////////////////////////////////////////////////
//
// Min(a, b) : return the minimum of a and b
//

template <class TYPE>
inline TYPE Min(TYPE a, TYPE b)
{
  if (a < b)
    return a;
  return b; 
}

/////////////////////////////////////////////////////////////////////////////
//
// Cube(a) : return the cube of a
//

template <class TYPE>
inline TYPE Cube(TYPE f)
{
  return f * f * f;
}


/////////////////////////////////////////////////////////////////////////////
//
// Clamp(x, min, max) : clip x to be in the interval [min, max]
//

template <class TYPE>
inline TYPE Clamp(TYPE x, TYPE min, TYPE max)
{
  if (x < min)
    return min;
  else if (x > max)
    return max;
  return x; 
}

/////////////////////////////////////////////////////////////////////////////
//
// ZoomToDegrees(zoom) : return the degrees variant of given zoom
//

template <class TYPE>
inline TYPE ZoomToDegrees(TYPE z)
{
  return RADTODEG*2.0*atan(1.0/(z));
}

/////////////////////////////////////////////////////////////////////////////
//
// ZoomToDegrees(degrees) : return the zoom variant of given degrees
//

template <class TYPE>
inline TYPE DegreesToZoom(TYPE d)
{
  return 1.0/tan(0.5*DEGTORAD*(d));
}

/////////////////////////////////////////////////////////////////////////////
//
// Bit Flag functions
//
//  useful for short, int, etc.
//
/////////////////////////////////////////////////////////////////////////////
//
// RaiseBitFlag(var, FLAG) : raise (set) the bit flag FLAG in the given var
//

template <class TYPE>
inline void RaiseBitFlag(TYPE &var, TYPE flag)
{
  var |= flag;
}

/////////////////////////////////////////////////////////////////////////////
//
// LowerBitFlag(var, FLAG) : lower the bit flag FLAG in the given var
//

template <class TYPE>
inline void LowerBitFlag(TYPE &var, TYPE flag)
{
  var &= ~flag;
}

/////////////////////////////////////////////////////////////////////////////
//
// BitFlagRaised(var, FLAG) : is the bit flag FLAG set (raised) ?
//

template <class TYPE>
inline TYPE BitFlagRaised(TYPE var, TYPE flag)
{
  return var & flag;
}

/////////////////////////////////////////////////////////////////////////////
//
// BitFlagLowered(var, FLAG) : is the bit flag FLAG lowered ?
//

template <class TYPE>
inline TYPE BitFlagLowered(TYPE var, TYPE flag)
{
  return !(var & flag);
}


/////////////////////////////////////////////////////////////////////////////
//
// Utility functions 
//
// some taken from libgm 
//
/////////////////////////////////////////////////////////////////////////////
//
// Ceil
//

inline double Ceil(double f)
{
  return (f == int(f)) ? f : (f > 0) ? double(int(f) + 1) : double(int(f));
}

/////////////////////////////////////////////////////////////////////////////
//
// Degrees : convert given radians to degrees
//

inline double Degrees(double f)
{
  return f * RADTODEG;
}

/////////////////////////////////////////////////////////////////////////////
//
// Floor
//

inline double Floor(double f)
{
  return (f == int(f)) ? f : (f > 0) ? double(int(f)) : double(int(f) - 1);
}

/////////////////////////////////////////////////////////////////////////////
//
// Inv
//

inline double Inv(double f)
{
  return 1.0 / f;
}

/////////////////////////////////////////////////////////////////////////////
//
// IsZero
//

inline bool IsZero(double f)
{
  return (f < EPSILON && f > -EPSILON);
}

/////////////////////////////////////////////////////////////////////////////
//
// Radians : convert given degrees to radians 
//

inline double Radians(double f)
{
  return f * DEGTORAD;
}

/////////////////////////////////////////////////////////////////////////////
//
// Round
//

inline double Round(double f)
{
  return (f >= 0) ? double(int(f + 0.5)) : double(- int(0.5 - f));
}

/////////////////////////////////////////////////////////////////////////////
//
// Sign
//

inline double Sign(double f)
{
  return (f < 0) ? -1.0 : 1.0;
}

/////////////////////////////////////////////////////////////////////////////
//
// Smooth
//

inline double Smooth(double f)
{
  return (3.0 - 2.0 * f) * f * f;
}

/////////////////////////////////////////////////////////////////////////////
//
// Trunc
//

inline double Trunc(double f)
{
  return double(int(f));
}

/////////////////////////////////////////////////////////////////////////////
//
// ZSign
//

inline double ZSign(double f)
{
  return (f > 0) ? 1.0 : (f < 0) ? -1.0 : 0.0;
}

/////////////////////////////////////////////////////////////////////////////
//
// FuzEQ
//

//
// Johan, EPSILON is ***MUCH*** too small for use with floats.
// It caused the alarm about illegal normals.
// I did not dare to change it, so I added another
// macro, but your sim code may benefit from a much
// bigger EPSILON. I recommend 10E-5 instead of 10E-10
// Or maybe define two: one for floats, one for doubles.
//  - Bram.
//

inline bool FuzEQ(double f, double g)
{
  return (f <= g) ? (f >= g - EPSILON) : (f <= g + EPSILON);
}

inline bool FuzEQ(float x, float y, float tolerance=0.00001)
{
  return (fabs((x)-(y)) < 0.00001);
}

/////////////////////////////////////////////////////////////////////////////
//
// FuzGEQ
//

inline bool FuzGEQ(double f, double g)
{
  return (f >= g - EPSILON);
}

/////////////////////////////////////////////////////////////////////////////
//
// FuzLEQ
//

inline bool FuzLEQ(double f, double g)
{
  return (f <= g + EPSILON);
}

/////////////////////////////////////////////////////////////////////////////
//
// Lerp : linearly interpolate 
//

inline double Lerp(double f, double l, double h)
{
  return l + ((h - l) * f );
}

/////////////////////////////////////////////////////////////////////////////
//
// Slide
//

inline double Slide(double f, double l, double h)
{
  return (f < 0) ? l : (f > 1) ? h : Lerp(Smooth(f), l, h);
}


//
// AlignValue
//  align given value (number) to given alignment
//  ZUCHT! ptrs and integers do NOT have the same size!
//  This bug took me an entire night! - Bram
//

inline long AlignValue
(
  long    val,
  long    alignment_size_in_bytes
)
{
  long bits = alignment_size_in_bytes - 1;
  return (val + bits) & ~bits;
}


inline int AlignValue
(
  int     val,
  int     alignment_size_in_bytes
)
{
  int bits = alignment_size_in_bytes - 1;
  return (val + bits) & ~bits;
}

//
// ClipValue
//  align given value to interval [lo, hi]
//

template <class TYPE>
inline TYPE ClipValue(const TYPE &val, 
					  const TYPE &lo,
					  const TYPE &hi)
{
  return (val < lo ? lo: val > hi ? hi: val);
}
		       
#endif

