/*
 * (c) 1996, 1997, 1998 by Polar Pyramid.
 */

// Description:
//	Inline member functions of the Vector2 class
//
// History:
//	Oct 31, 1996:	creation JSt
//
//

#ifndef VECTOR2_INL
#define VECTOR2_INL

#include <assert.h>
#include "auxil.h"

#include "vector2.h"

/////////////////////////////////////////////////////////////////////////////
//
// Default Constructor
//

inline Vector2::Vector2()
 : x(0), y(0)
{
}

/////////////////////////////////////////////////////////////////////////////
//
// Copy Constructor
//

#ifdef WHY_IMPLEMENT_BUILDINS
inline Vector2::Vector2(const Vector2& v)
  : x(v.x), y(v.y)
{
}
#endif

inline Vector2::Vector2(const Vector3& v)
  : x(v.x), y(v.y)
{
}

/////////////////////////////////////////////////////////////////////////////
//
// Element Constructor
//

inline Vector2::Vector2(float ix, float iy)
 : x(ix), y(iy)
{
}

/////////////////////////////////////////////////////////////////////////////
//
// Array Access
//

inline float& Vector2::operator [](int i)
{
  switch (i)	
  {
  case 0: return x;
  case 1: return y;
  }
  ASSERT(false);
  return x;
}

inline const float& Vector2::operator [](int i) const
{
  switch (i)	
  {
  case 0: return x;
  case 1: return y;
  }
  ASSERT(false);
  return x;
}

/////////////////////////////////////////////////////////////////////////////
//
// Assignment
//

inline Vector2& Vector2::assign(float ix, float iy)
{
  x = ix;
  y = iy;
  return *this;
}

#ifdef WHY_IMPLEMENT_BUILDINS
// assignment operator slows things down considerably !!
// e.g. a = b + c; is then much slower, so we do not define this operator
/////////////////////////////////////////////////////////////////////////////
//
// Vector2 = Vector2
//

inline Vector2& Vector2::operator =(const Vector2& v)
{
  x = v.x; 
  y = v.y;
  return *this;
}
#endif

/////////////////////////////////////////////////////////////////////////////
// Math Operators
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
//
// - Vector2
//

inline Vector2 Vector2::operator -() const
{
  return Vector2(-x, -y);
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector2 += Vector2
//

inline Vector2& Vector2::operator += (const Vector2& v)
{
  x += v.x; 
  y += v.y;
  return *this;
}


/////////////////////////////////////////////////////////////////////////////
//
// Vector2 -= Vector2
//


inline Vector2& Vector2::operator -= (const Vector2& v)
{
  x -= v.x; 
  y -= v.y;
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector2 *= float
//


inline Vector2& Vector2::operator *= (float c)
{
  x *= c; 
  y *= c;
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector2 /= float
//


inline Vector2& Vector2::operator /= (float c)
{
  float inv = 1.0f/c;
  x *= inv; 
  y *= inv;
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector2 + Vector2
//


inline Vector2 Vector2::operator + (const Vector2& v) const
{
  return Vector2(x + v.x, 
		 y + v.y);
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector2 - Vector2
//


inline Vector2 Vector2::operator - (const Vector2& v) const
{
  return Vector2(x - v.x, 
		 y - v.y);
}


/////////////////////////////////////////////////////////////////////////////
//
// Vector2 / float
// 


inline Vector2 Vector2::operator / (float c) const
{
  return Vector2(x/c, 
		 y/c);
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector2 == Vector2
//


inline bool Vector2::operator == (const Vector2& v) const
{
  return ((x == v.x) && 
          (y == v.y));
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector2 != Vector2
//


inline bool Vector2::operator != (const Vector2& v) const
{
  return (!(*this == v));
}

/////////////////////////////////////////////////////////////////////////////
//
// float * Vector2
//


inline Vector2 operator * (float c, Vector2 const &v)
{
  return Vector2(v.x*c, 
		 v.y*c);
}

/////////////////////////////////////////////////////////////////////////////
//
// Vector2 * float
//


inline Vector2 operator * (Vector2 const &v, float c)
{
  return Vector2(v.x*c, 
		 v.y*c);
}

/////////////////////////////////////////////////////////////////////////////
//
// copyTo
//


inline void Vector2::copyTo(float f[2]) const
{
  f[0] = x; 
  f[1] = y;
}

inline void Vector2::Dump(void) const
{
  auxil_log(LOG_INF, "[%g %g]", x, y);
}

/////////////////////////////////////////////////////////////////////////////
//
// length : returns the length of the vector2 as a double, note that 
//          this has to return a double because it uses the sqrt function
//

inline double Vector2::length() const
{
  return sqrt(lengthSqr());
}

/////////////////////////////////////////////////////////////////////////////
//
// lengthSqr : returns the squared length of the vector2
//


inline float Vector2::lengthSqr() const
{
  return x*x + y*y;
}

/////////////////////////////////////////////////////////////////////////////
//
// normalize : will normalize the vector2, make it of length 1
//

inline void Vector2::normalize(void)
{
  double len = length();
  if (len != 0.0)      // avoid SIGFPE
    *this *= 1.0f/len;
}

/////////////////////////////////////////////////////////////////////////////
//
// dot : returns the dot product of 2 vector2s
//

inline float Vector2::dot(const Vector2& v2) const
{
  return x*v2.x + y*v2.y;
}

/////////////////////////////////////////////////////////////////////////////
//
// distanceSqr : return the squared distance between 2 given points which 
//               avoids taking an expensive sqrt call. Usefull when comparing
//               different distances.
//

inline float Vector2::distanceSqr(const Vector2& v2) const
{
  return (Sqr(x - v2.x) +
	  Sqr(y - v2.y));
}

/////////////////////////////////////////////////////////////////////////////
//
// distance : return the distance between 2 given points, note that 
//            this has to return a double because it uses the sqrt function
//

inline double Vector2::distance(const Vector2& v2) const
{
  return sqrt(distanceSqr(v2));
}

inline void Vector2::Snap(float grid_spacing)
{
  x = SnapValue(x, grid_spacing);
  y = SnapValue(y, grid_spacing);
}

#endif
