/*
 * bbox.cc
 *
 * (c)96 Bram Stolk
 *
 * 960504 Hacked bounding-box calculation.
 * 960506 After _major_ debugging sessions, and many problems,
 *        I finaly found out that our vscreen is too small. The height of
 *        the heighest bounding box (267) is larger than the overlap of
 *        two neighbouring screens (256). Therefore I have decided to increase
 *        the size of VSCREEN.
 *        This will lower the performance of the rot code, due to misses
 *        in the data-cache of the Pentium (Don't confuse with the misses
 *        in our own virtual-screen caches.)
 * 960528 I made virtual.c now almost redundant.
 *        We will now use primitives that are copied to form a rotation
 *        buffer. This is done each frame. We only need to keep the 
 *        bouding box calculation. 
 * 960603 Renamed to bbox.c, moved to subsys rot
 * 961020 Changed to C++ JSt
 * 961109 Used auxil_nbdir() - BSt
 *   
 */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "setup.h"
#include "auxil.h"
#include "rot.h"

/*
 * This specifies bounding boxes around the area in the
 * virtual screen that is copied to the gfx buffer.
 * This bounding box is different for each rotation value.
 * 0 degrees rot has bounding box equal to the playing screen.
 * 90 degrees has bounding box with w/h reversed.
 * 45 degrees has a very large bounding box.
 */
struct bbox {
  signed short minx;
  signed short maxx;
  signed short miny;
  signed short maxy;
  signed short w,h;
};

static struct bbox boundingboxes[ROTSTEPS];
static int bbox_maxw=0, bbox_maxh=0;

/*---------------------------------------------------------------------------*/

/*
 * For all possible rotations of the playing screen, we
 * calculate its bounding box. For rotation 0, the bounding
 * box ofcourse equals the playing screen itself.
 */

void rot_calculateboundingboxes() 
{
  int rot;
  int x,y;
  unsigned short *virbuf1=0;
  unsigned short *virbuf2=0;
  unsigned short *spritebuf=0;
  unsigned short *writer1=0;
  unsigned short *writer2=0;
  unsigned short *reader=0;
  unsigned short lo,hi;
  int offx = (ROTW/2);
  int offy = (ROTH/2);
  FILE *f;

  std::string fname = GetRootPath() + "/bbox.dat";

  f=fopen(fname.c_str(),"rb");
  if (f) 
  {
    if (fread(boundingboxes, sizeof(struct bbox), ROTSTEPS, f) != ROTSTEPS)
      auxil_log
      (
        LOG_ERR,
        "bbox.dat corrupt. Please throw away. I will remake!"
      );
    fclose(f);
    return;
  }
  
  virbuf1=(unsigned short *)calloc(1,ROTW*ROTH*2);
  virbuf2=(unsigned short *)calloc(1,ROTW*ROTH*2);
  spritebuf = (unsigned short *)calloc(1,SPRW*SPRH*2);
  assert(virbuf1 && virbuf2 && spritebuf);
  writer1 = virbuf1;
  writer2 = virbuf2;
  for (y=0; y<ROTH; y++) {
    for (x=0; x<ROTW; x++) {
      *writer1++ = (unsigned short) x;
      *writer2++ = (unsigned short) y;
    }
  }

  for (rot=0; rot<ROTSTEPS; rot++) {
    rot_rotscr(rot, (char *)(virbuf1+ROTW*offy+offx), (char *)(spritebuf+SPROFF));
    /* search lo/hi vals */
    lo=1000; hi=0;
    for (y=0; y<PSCREENH; y++) {
      reader = spritebuf + SPROFF + y*SPRW;
      for (x=0; x<PSCREENW; x++) {
        if (*reader<lo) lo = *reader;
        if (*reader>hi) hi = *reader;
        reader++;
      }
    }
    boundingboxes[rot].minx = lo - offx;
    boundingboxes[rot].maxx = hi - offx;
    boundingboxes[rot].w = hi - lo;
    if (boundingboxes[rot].w > bbox_maxw)
      bbox_maxw = boundingboxes[rot].w;

    rot_rotscr(rot, (char *)(virbuf2+ROTW*offy+offx), (char *)(spritebuf+SPROFF));
    /* search lo/hi vals */
    lo=1000; hi=0;
    for (y=0; y<PSCREENH; y++) {
      reader = spritebuf + SPROFF + y*SPRW;
      for (x=0; x<PSCREENW; x++) {
        if (*reader<lo) lo = *reader;
        if (*reader>hi) hi = *reader;
        reader++;
      }
    }
    boundingboxes[rot].miny = lo - offy;
    boundingboxes[rot].maxy = hi - offy;
    boundingboxes[rot].h = hi - lo;
    if (boundingboxes[rot].h > bbox_maxh)
      bbox_maxh = boundingboxes[rot].h;
  }

  auxil_log(LOG_INF,"widest bounding box is %d pixels wide",bbox_maxw);
  auxil_log(LOG_INF,"highest bounding box is %d pixels high",bbox_maxh);
  free(virbuf1);
  free(virbuf2);
  free(spritebuf);

  f=fopen(fname.c_str(),"wb");
  if (!f) return;
  if (fwrite(boundingboxes, sizeof(struct bbox), ROTSTEPS, f) != ROTSTEPS)
    auxil_log(LOG_FAT,"couldnt write bbox.dat");
  fclose(f);
}

/*---------------------------------------------------------------------------*/

/*
 * For a given rotation and player position, we calculate a bounding
 * box of the exposed world data.
 */

void rot_getbbox(int rot, int px, int py, int *x, int *y, int *w, int *h) 
{
  struct bbox *bb;

  bb = boundingboxes+rot;
  *x = px + bb->minx;
  *y = py + bb->miny;
  *w = 1 + (bb->maxx - bb->minx); /* why not store width & height ? (Johan)*/
  *h = 1 + (bb->maxy - bb->miny); /* (max is then not needed) */
}

/*---------------------------------------------------------------------------*/

