/*
 * Tux vs Clippy
 * (c)2002-2003 by Bram Stolk
 * b.stolk at chello.nl
 * Licensed under GPL
 */

//
// screen.cc
//
// implements class GfxScreen
//
//    --------   A = sprite buffer
//   |A _____|   B = visible sprite buffer
//   | |B    |   (extra space in A is clipping space)
//   | |     |
//   |  -----|
//    --------
//
// 020911 BSt
//   Ported to C++ and xbox
// 970201 BSt
//   Initial version
// JSt 970207
//   Bug fixed in ClearArea 
// 970215 JSt
//   Bug fixed in DrawImage
// 970216 JSt
//   Added DrawTractorBeam, GetScreenPos
// 970218 FS
//   Inlined extended_copy(), seems to work
// 970220 JSt
//   Added DrawLine, InterlaceModes, 
//   fixed another bug in ClearArea (PSCREENPOSX & Y)
// 970227 JSt
//   Added DrawDxeLine, DrawDxeShadowLine, 
//   will crash when compiled optimized ...
// 970420 JSt
//   Added DrawRawImage
// 970610 BSt
//   Added DrawScreenSpaceText
// 970914 JSt
//   Added DrawPixel()
// 971015 JSt
//   Added DrawScreenSpaceText, split old version into text & box parts
//

#include "setup.h"
#include "config.h"
#include "auxil.h"
#include "sprite.h"
#include "im16.h"
#include "maths.h"

#include "rot.h"

#include "gfxscreen.h"


//
// Constructor
//

GfxScreen::GfxScreen(void) :
  spr_im(0),
  sprite_buf(0),
  visible_sprite_buf(0),
  pivot_position(0.0, 0.0, 0.0),
  pivot_discreterotation(0)
{
  spr_im = new Image16(SPRW, SPRH);

  sprite_buf = spr_im->data;
  visible_sprite_buf = sprite_buf + SPROFF;

  rot_init();
}

//
// Destructor
//

GfxScreen::~GfxScreen(void) 
{
  delete spr_im;
}


void GfxScreen::DrawScreenSpaceSprite
(
  const Sprite   &sprite,
  const Vector3  &pos,         // position of this sprite
  bool           is_oriented,
  float          parm
) const
{
  int screenp[2];
  float angle;

  if (is_oriented)
  {
    angle = parm;
    ValidateAngle(parm); // Bug fix JSt 970919
  }

  screenp[0] = (int) pos[0];
  screenp[1] = (int) pos[1];

#if 0
  // correct sprite position

  screenp[0] -= sprite.w/2;
  screenp[1] -= sprite.h/2;
#endif

  // clipping 

  if (screenp[0] > -sprite.w &&
      screenp[0] < PSCREENW &&
      screenp[1] > -sprite.h &&
      screenp[1] < PSCREENH )
  {
    // calc. starting frame according to pivot rotation

    int integer_fr;
    if (is_oriented)
    {
      float fr = ((ROTSTEPS-1.0) - pivot_discreterotation) * sprite.framecount;
      fr *= 1.0/128.0;

      // calc. actual frame according to sprite rotation
  
      fr += (((angle + 1.5*EENPI)/(TWEEPI)) * sprite.framecount);
  
      integer_fr = (int) (fr + 0.5);
      integer_fr %= sprite.framecount;
    }
    else
    {
      integer_fr = (int) (parm * sprite.framecount);
      if (integer_fr >= (int)sprite.framecount)
        integer_fr = sprite.framecount-1;
    }

    assert(integer_fr>=0);
    assert(integer_fr<(int)sprite.framecount);
    sprite.Render(integer_fr, screenp[0], screenp[1], visible_sprite_buf);
  }
  else
  {
  }
}


void GfxScreen::DrawSprite
(
  const Sprite   &sprite,
  const Vector3  &pos,         // position of this sprite
  bool           is_oriented,
  float          parm
) const
{
  int screenp[2];
  float angle;

  if (is_oriented)
  {
    angle = parm;
    ValidateAngle(parm); // Bug fix JSt 970919
  }

  GetScreenPos(pos, screenp);

  // correct sprite position

  screenp[0] -= sprite.w/2;
  screenp[1] -= sprite.h/2;

  // clipping 

  if (screenp[0] > -sprite.w &&
      screenp[0] < PSCREENW &&
      screenp[1] > -sprite.h &&
      screenp[1] < PSCREENH )
  {
    // calc. starting frame according to pivot rotation

    int integer_fr;
    if (is_oriented)
    {
      float fr = ((ROTSTEPS-1.0) - pivot_discreterotation) * sprite.framecount;
      fr *= 1.0/128.0;

      // calc. actual frame according to sprite rotation
  
      fr += (((angle + 1.5*EENPI)/(TWEEPI)) * sprite.framecount);
  
      integer_fr = (int) (fr + 0.5);
      integer_fr %= sprite.framecount;
    }
    else
    {
      integer_fr = (int) (parm * sprite.framecount);
      if (integer_fr >= (int)sprite.framecount)
        integer_fr = sprite.framecount-1;
    }

    assert(integer_fr>=0);
    assert(integer_fr<(int)sprite.framecount);
    sprite.Render(integer_fr, screenp[0], screenp[1], visible_sprite_buf);
  }
  else
  {
  }
}

//
// SetPivot
//

void GfxScreen::SetPivot(const Vector3 &pos, float heading)
{
  pivot_position = pos;
  ValidateAngle(heading);
  pivot_discreterotation = (unsigned char) (heading * ROTSTEPS / TWEEPI);
  if (pivot_discreterotation==ROTSTEPS)
    pivot_discreterotation--;
}


void GfxScreen::RotateBuffer(short *rotbuf)
{
  assert(pivot_discreterotation < ROTSTEPS);
  int offx = (ROTW/2);
  int offy = (ROTH/2);

  rot_rotscr
  (
    pivot_discreterotation,
    (char *) (rotbuf+ROTW*offy+offx),
    (char *) GetVisibleSpriteBuffer()
  );
}


Vector2 GfxScreen::VectorToScreenSpace(const Vector2 &v) const
{
  Vector2 retval;
  RotateVector2((ROTSTEPS-1-pivot_discreterotation)*2, v, retval);
  return retval;
}


void GfxScreen::GetScreenPos
(
  const Vector3 &world_pos,
  int            screen_pos[2]
) const
{
  assert(screen_pos != 0);
  Vector2 locp;
  Vector2 dif(world_pos - pivot_position);
  RotateVector2(pivot_discreterotation*2, dif, locp);
  screen_pos[0] = (int) (locp.x + POS_CENTRE_X);
  screen_pos[1] = (int) (locp.y + POS_CENTRE_Y);
}


GfxScreen &GfxScreen::TheOne(void)
{
  static GfxScreen *scn = new GfxScreen();
  return *scn;
}

