/*
 * Tux vs Clippy
 * (c)2002-2003 by Bram Stolk
 * b.stolk at chello.nl
 * Licensed under GPL
 */

#include <stdio.h>
#include <assert.h>
#include <string.h>

#define WITHTIFF

#ifdef WITHTIFF
#include <tiff.h>
#include <tiffio.h>
#endif

#include "auxil.h"

#include "im8.h"
#include "im16.h"
#include "im32.h"

#include "colconv.h"

#define SWAP_RED_BLUE


Image32::Image32(const std::string &fname) :
  w(0),
  h(0),
  maxval(255),
  data(0),
  has_alpha(false),
  alpha_is_premultiplied(false)
{
  if (strstr(fname.c_str(),".ppm"))
  {
    ReadPPM(fname.c_str());
    return;
  }
  if (strstr(fname.c_str(),".tif"))
  {
    ReadTIFF(fname.c_str());
    return;
  }
  assert(0);
}


void Image32::ReadPPM(const char *fname)
{
  FILE *f=fopen(fname,"rb");
  if (!f)
    auxil_log(LOG_ERR,"Cannot open file '%s' for reading", fname);
  assert(f);
  int retval = fscanf(f, "P6 %d %d %d ", &w, &h, &maxval);
  assert(retval==3);
  assert(maxval==255);
  unsigned char tmp[w*h*3];
  data = new unsigned char[w*h*4];
  retval = fread(tmp, w*3, h, f);
  assert(retval == h);
  fclose(f);
  unsigned char *reader = tmp;
  unsigned char *writer = data;
  for (int i=0; i<w*h; i++)
  {
    *writer++ = reader[0];
    *writer++ = reader[1];
    *writer++ = reader[2];
    reader+=3;
    *writer++ = 255;
  }
}


void Image32::ReadTIFF(const char *fname)
{
#ifdef WITHTIFF
  TIFFSetWarningHandler(0);
  TIFF *tif = TIFFOpen(fname,"r");
  assert(tif);

  TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
  TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);

  data = new unsigned char[w*h*4];

  uint16 extracount=0;
  uint16 *extratypes;

  auxil_log(LOG_INF, "Reading TIFF file '%s' of size %dx%d",fname,w,h);
  int retval=TIFFGetField(tif, TIFFTAG_EXTRASAMPLES, &extracount, &extratypes);
  if (retval)
  {
    auxil_log(LOG_INF,"tiff has %d extra samples", extracount);
    if (extracount>0)
    {
      switch(extratypes[0])
      {
        case 1:
          has_alpha = true;
          alpha_is_premultiplied = true;
          break;
        case 2:
          has_alpha = true;
          alpha_is_premultiplied = false;
          break;
        default:
          printf
          (
            "Unknown extra_sample type %d in tif\n",
            extratypes[0]
          );
      }
    }
  }

  int tmp[h][w];
  // Do the reading.
  if (!TIFFReadRGBAImage(tif, w, h, (uint32 *)(tmp), 0))
  {
    printf("Error reading rgba-image\n");
  }
  TIFFClose(tif);

  for (int y=0; y<h; y++)
  {
    memcpy(data+y*w*4, tmp[h-y-1], w*4);
  }
#else
  assert(0);
#endif
}


Image32::Image32(int ww, int hh) :
  w(ww),
  h(hh),
  maxval(255),
  data(0),
  has_alpha(false),
  alpha_is_premultiplied(false)
{
  data = new unsigned char[4*w*h];
}


Image32::Image32(const Image16 &im16) :
  w(im16.w),
  h(im16.h),
  maxval(255),
  data(0)
{
  data = new unsigned char[w*h*4];
  short *reader = im16.data;
  unsigned char  *writer = data;
  for (int y=0; y<h; y++)
  {
    for (int x=0; x<w; x++)
    {
      *writer++ = COL2R((*reader));
      *writer++ = COL2G((*reader));
      *writer++ = COL2B((*reader));
      *writer++ = 255;
      reader++;
    }
  }
}


void Image32::Write(const std::string &fname) const
{
  FILE *f=fopen(fname.c_str(),"wb");
  if (!f)
  {
    auxil_log(LOG_ERR,"Cannot write to file '%s'", fname.c_str());
    return;
  }
  fprintf(f,"P6\n%d %d\n%d\n",w,h,maxval);
  unsigned char tmp[w*h*3];
  unsigned char *reader = data;
  unsigned char *writer = tmp;
  for (int i=0; i<w*h; i++)
  {
    *writer++ = *reader++;
    *writer++ = *reader++;
    *writer++ = *reader++;
    reader++;
  }
  int retval = fwrite(tmp, 3*w, h, f);
  assert(retval == h);
  fclose(f);
}


void Image32::SwapRedBlue(void)
{
  unsigned char *p = data;
  unsigned char tmp;
  for (int i=0; i<w*h; i++)
  {
    tmp=p[0];
    p[0]=p[2];
    p[2]=tmp;
    p+=4;
  }
}


Image32::~Image32() 
{
  delete data; 
}


void Image32::ChromaKey(int r, int g, int b)
{
  int a0=0,a1=0;
  unsigned char *reader=data;
  for (int i=0; i<w*h; i++)
  {
    if (reader[0]==r && reader[1]==g && reader[2]==b)
    {
      a0++;
      reader[3]=0;
    }
    else
    {
      a1++;
      reader[3]=255;
    }
    reader+=4;
  }
  has_alpha = true;
  alpha_is_premultiplied = false;
  printf
  (
    "%.1f %% of the pixels have been keyed\n",
    (100.0*a0) / (a0+a1)
  );
}


Image8 *Image32::GetComponent(const std::string &comp) const
{
  int offs = -1;
  if (comp=="red") offs=0;
  if (comp=="green") offs=1;
  if (comp=="blue") offs=2;
  if (comp=="alpha") offs=3;
  assert(offs!=-1);
  Image8 *retval = new Image8(w,h);
  unsigned char *reader = data;
  for (int i=0; i<w*h; i++)
  {
    retval->data[i] = reader[offs];
    reader+=4;
  }
  return retval;
}


void Image32::SetComponent(const std::string &comp, const Image8 &im8)
{
  assert(w == im8.w);
  assert(h == im8.h);
  int offs = -1;
  if (comp=="red") offs=0;
  if (comp=="green") offs=1;
  if (comp=="blue") offs=2;
  if (comp=="alpha") offs=3;
  assert(offs!=-1);
  for (int i=0; i<w*h; i++)
    data[i*4+offs] = im8.data[i];
}


