/*
 * Tux vs Clippy
 * (c)2002-2003 by Bram Stolk
 * b.stolk at chello.nl
 * Licensed under GPL
 */

#include <assert.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>     // for pthread_kill
#include <sys/ioctl.h>
#include <fcntl.h>
#include <linux/soundcard.h>

#include "auxil.h"
#include "soundclip.h"
#include "soundengineoss.h"


/*
 * Set sound device parameters to given values. Return -1 if
 * values not valid. Sampling rate is returned.
 */
static int set_dsp_params(int fd, int channels, int bits, int *rate) 
{
  int status, val = channels;

  status = ioctl(fd, SOUND_PCM_WRITE_CHANNELS, &val);
  if (status == -1)
    perror("SOUND_PCM_WRITE_CHANNELS ioctl failed");
  if (val != channels) /* not valid, so return */
    return -1;
  val = bits;
  status = ioctl(fd, SOUND_PCM_WRITE_BITS, &val);
  if (status == -1)
    perror("SOUND_PCM_WRITE_BITS ioctl failed");
  if (val != bits)
    return -1;
  status = ioctl(fd, SOUND_PCM_WRITE_RATE, rate);
  if (status == -1)
    perror("SOUND_PCM_WRITE_RATE ioctl failed");
  return 0;
}



void *SoundEngineOSS::UpdateThread(void *job)
{
  puts("UPDATE THREAD!");
  SoundEngineOSS *se = static_cast<SoundEngineOSS *>(job);
  assert(se);
  while (1)
  {
    se->Sustain();
    usleep(10000); // limit to 100Hz update
  }
}



SoundEngineOSS::SoundEngineOSS() :
fd(0),
bytes_written(0),
semaphore(false)
{
  int retval;
  puts("CREATE SOUNDENGINE!");

  char *device = "/dev/dsp";

  /* try to open device */
  fd = open(device, O_RDWR);
  if (fd == -1) 
  {
    perror("open /dev/dsp failed");
    auxil_log(LOG_ERR,"Cannot create OSS soundengine");
    return;
  }

  assert(fd!=-1);

  int version;
  int status = ioctl(fd, OSS_GETVERSION, &version);
  if (status ==  -1)
    perror("OSS_GETVERSION ioctl failed");
  printf("OSS Version %x\n", version);

  int rate=45000;
  set_dsp_params(fd, 2, 16, &rate);

  opened = true;

  PlayDirect("fire");

#if 1
  thread_id=0;
  retval = pthread_create
  (
    &thread_id,
    0, // default attribs
    (void *(*)(void *)) UpdateThread,
    (void *) this
  );
  if (retval)
  {
    perror("pthread_create() failed");
    auxil_log(LOG_FAT,"No snd thread");
  }
  else
  {
    printf("threadid = %d\n", thread_id);
  }

  retval = pthread_detach(thread_id);
  if (retval)
    perror("pthread_detach() failed");
#endif

  // give other thread some time to clear the buffer
  usleep(100000);

  auxil_log(LOG_INF,"SoundEngineOSS created");
}


SoundEngineOSS::~SoundEngineOSS()
{
  if (!opened)
    return;

  int retval;

  retval=pthread_kill(thread_id, 9);
  if (retval==-1)
    perror("pthread_kill() failed");

  retval=close(fd);
  if (retval==-1)
    perror("close failed on /dev/dsp");
}


void SoundEngineOSS::Play(const std::string &name, int channr, int delay)
{
  assert(channr<SOUNDCHANNELCOUNT);
  if (!opened)
    return;
  SoundClip *clip = SoundClip::Get(name);
  assert(clip);
  GetSemaphore();
  channels[channr].Paste(clip, delay);
  ReleaseSemaphore();
}


void SoundEngineOSS::GetSemaphore(void)
{
  while (semaphore)
    usleep(10);
  semaphore = true;
}


void SoundEngineOSS::ReleaseSemaphore(void)
{
  assert(semaphore);
  semaphore = false;
}


void SoundEngineOSS::Send(void)
{
//  puts("SEND");
  SoundPacket *packets[SOUNDCHANNELCOUNT];
  int i,j;
  int cnt=0;
  for (i=0; i<SOUNDCHANNELCOUNT; i++)
  {
    packets[i] = channels[i].Pop();
    if (packets[i])
      cnt++;
  }
//  if (!cnt) return;

  SoundPacket out;
  for (j=0; j<SOUNDPACKETSIZE; j++)
  {
    short v=0;
    for (i=0; i<SOUNDCHANNELCOUNT; i++)
      if (packets[i])
        v += packets[i]->data[j];
    out.data[j] = v;
  }

  int rv = write(fd, out.data, SOUNDPACKETSIZE*sizeof(short));
  bytes_written += SOUNDPACKETSIZE*sizeof(short);
  if (rv==-1)
    perror("write() failed on /dev/dsp");

  for (i=0; i<SOUNDCHANNELCOUNT; i++)
    delete packets[i];
}


void SoundEngineOSS::PlayDirect(const std::string &name)
{
  assert(opened);
  SoundClip *clip = SoundClip::Get(name);
  int rv = write(fd, clip->data, clip->size*sizeof(short));
  bytes_written += clip->size*sizeof(short);
  if (rv==-1)
    perror("write() on /dev/dsp failed");
}


int SoundEngineOSS::GetOutputPointer(void)
{
  assert(opened);
  struct count_info cinfo_out;
  int rv = ioctl(fd, SNDCTL_DSP_GETOPTR, &cinfo_out);
  if (rv==-1)
    perror("SNDCTL_DSP_GETOPTR failed");
  return cinfo_out.bytes;
}


int SoundEngineOSS::GetInputPointer(void)
{
  assert(opened);
  struct count_info cinfo_in;
  int rv = ioctl(fd, SNDCTL_DSP_GETIPTR, &cinfo_in);
  if (rv==-1)
    perror("SNDCTL_DSP_GETIPTR failed");
  return cinfo_in.bytes;
}


// In update thread!
void SoundEngineOSS::Sustain()
{
  assert(opened);
  int retval;
#if 0
  // Only in OSS 3.9.4 and later :-(
  audio_errinfo errinfo;
  retval = ioctl(fd, SNDCTL_DSP_GETERROR, &errinfo);
  if (retval)
    perror("SNDCTL_DSP_GETERROR failed!");
  assert(retval!=-1);
  if (errinfo.play_underruns)
    auxil_log(LOG_ERR,"Audio underrun occurred");
#endif

  int delay=-1;
  retval = ioctl(fd, SNDCTL_DSP_GETODELAY, &delay);
  if (retval == -1)
    perror("SNDCTL_DSP_GETODELAY failed!");
  assert(retval != -1);
  //printf("delay=%d\n", delay);
  if (delay==0)
    auxil_log(LOG_ERR,"Audio underrun detected");
  if (delay > 0 && delay < SOUNDPACKETSIZE*sizeof(short))
  {
    GetSemaphore();
    Send();
    ReleaseSemaphore();
  }
}

