/*
 * Tux vs Clippy
 * (c)2002-2003 by Bram Stolk
 * b.stolk at chello.nl
 * Licensed under GPL
 */

#include <stdio.h>
#include <math.h>


#include "auxil.h"
#include "maths.h"
#include "im16.h"
#include "im32.h"
#include "wldprim.h"
#include "setup.h"
#include "world.h"



World::World(const std::string &tn) :
tilename(tn),
rot_im(ROTW, ROTH)
{
}


World::~World()
{
  cellmap_t::iterator ip;

  for (ip=cells.begin(); ip!=cells.end(); ++ip)
    delete (*ip).second;
  cells.clear();
}


void World::CreateWorld(const std::string &mapname)
{
  for (int i=0; i<16; i++)
  {
    char nm[80];
    sprintf
    (
      nm,
      "%s%c%c%c%c", 
      tilename.c_str(),
      (i&8)?'1':'0',
      (i&4)?'1':'0',
      (i&2)?'1':'0',
      (i&1)?'1':'0'
    );
    prims.Add(nm, i);
  }

  ReadMap(GetWorldPath()+mapname+".tif");
};



const WorldCell &World::AddCell(int x, int y, const std::string &primname)
{
  short lo = (short) x;
  short hi = (short) y;
  int idx = (hi<<16) + lo;
  WorldCell *cell = new WorldCell(x,y,primname,prims);
  cells[idx] = cell;
  return *cell;
}


const WorldCell *World::GetCell(int x, int y) const
{
  short lo = (short) x;
  short hi = (short) y;
  int idx = (hi<<16) + lo;
  cellmap_t::const_iterator ip = cells.find(idx);
  if (ip == cells.end())
    return 0;
  else
    return (*ip).second;
}


void World::ReadMap(const std::string &fname)
{
  Image32 im32(fname);
  Image16 im16(im32);

  int xs=im16.w-1;
  int ys=im16.h-1;

//  Image16 wldim(xs*128, ys*128);

  for (int y=0;y<im16.h; y++)
  {
    for (int x=0;x<im16.w; x++)
    {
      if (x>0 && y>0)
      {
        const std::string b0 = (im16.Pix(x-1,y-1))?"1":"0";
        const std::string b1 = (im16.Pix(x  ,y-1))?"1":"0";
        const std::string b2 = (im16.Pix(x-1,y  ))?"1":"0";
        const std::string b3 = (im16.Pix(x  ,y  ))?"1":"0";
        std::string primname = tilename+b3+b2+b1+b0;
        const WorldCell &cell = AddCell(x - (im16.w)/2, y - (im16.h)/2, primname);
//        wldim.Paste(*(cell.prim.im), (x-1)*128, (y-1)*128);
      }
    }
  }
//  wldim.Write("wld.ppm");
}


void World::RenderRotationBuffer(Vector3 &pos)
{
  rot_im.Clear(0x001f);

  int cx,cy;
  int xo,yo;

  if (pos[0]>=0)
    cx = (int) ((pos[0]+64.0) / 128.0f);
  else
    cx = (int) ((pos[0]-64.0) / 128.0f);

  if (pos[1]>=0)
    cy = (int) ((pos[1]+64.0) / 128.0f);
  else
    cy = (int) ((pos[1]-64.0) / 128.0f);

  xo=(int) (cx*128.0 - pos[0]);
  yo=(int) (cy*128.0 - pos[1]);


//  auxil_log(LOG_INF,"pos=(%f,%f), cx,cy=%d,%d, xo,yo=%d,%d",
//    pos[0],pos[1],cx,cy,xo,yo);

  int misses=0;
  for (int yy=-3; yy<=3; yy++)
  {
    for (int xx=-3; xx<=3; xx++)
    {
      const WorldCell *cell = GetCell(cx+xx, cy+yy);
      if (cell) 
      {
        int xoff = xx*128 + xo + ROTW/2 - 64;
        int yoff = yy*128 + yo + ROTH/2 - 64;
        rot_im.ClipAndPaste(*(cell->prim.im), xoff, yoff);
      }
      else
      {
        misses++;
      }
    }
  }
}


const Image16 &World::GetRotationBuffer(void) const
{
  return rot_im;
}


bool World::IsValidWorldPosition(const Vector3 &pos) const
{
  int cx,cy;
  int xo,yo;

  if (pos[0]>=0)
    cx = (int) ((pos[0]+64.0) / 128.0f);
  else
    cx = (int) ((pos[0]-64.0) / 128.0f);

  if (pos[1]>=0)
    cy = (int) ((pos[1]+64.0) / 128.0f);
  else
    cy = (int) ((pos[1]-64.0) / 128.0f);

  xo=(int) (cx*128.0 - pos[0]);
  yo=(int) (cy*128.0 - pos[1]);

//  printf("xo,yo=%d,%d\n", xo,yo);

  int msk;
  if (xo > 0)
  {
    if (yo > 0)
      msk=1;
    else
      msk=4;
  }
  else
  {
    if (yo > 0)
      msk=2;
    else
      msk=8;
  }
  const WorldCell *cell = GetCell(cx, cy);
  if (!cell) return false;
  bool valid = (cell->prim.footprint & msk);
  return valid;
}


Vector3 World::GetRandomWorldPosition(void) const
{
  while (1)
  {
    Vector3 rpos((rand()&1023)-512,(rand()&1023)-512,0);
    if (IsValidWorldPosition(rpos) && rpos.length()>300)
      return rpos;
  }
}

